# frozen_string_literal: true

require "is/extension"
require "is/localized"
require "is/stateful"

module Is
  # [public] Customized inspections for any object.
  #
  module Inspectable
    extend Is::Extension

    extends dependencies: [Is::Stateful]

    extends :definition do
      # [public] Sets one or more instance variables or instance methods as inspectable.
      #
      # without - An array of instance variables or instance methods that should not be inspected.
      #
      def inspects(*inspectables, without: [])
        if inspectables.any?
          mutate_state :__inspectables do |current_inspectables|
            current_inspectables.concat(inspectables.map(&:to_s)).uniq
          end
        end

        if without.any?
          mutate_state :__uninspectables do |current_uninspectables|
            current_uninspectables.concat(without.map(&:to_s)).uniq
          end
        end
      end
    end

    applies do
      state :__inspectables, default: []
      state :__uninspectables, default: []
    end

    # [public] Inspects the object.
    #
    def inspect
      inspection = +"#<#{self.class}:0x#{__id__.to_s(16)}"

      if Inspectable.inspecting?(self)
        "#{inspection} ...>"
      else
        Inspectable.prevent_recursion(self) do
          each_inspectable do |inspectable|
            value = if inspectable.start_with?("@")
              instance_variable_get(inspectable)
            else
              send(inspectable)
            end

            inspection << ", #{inspectable}=#{value.inspect}"
          end

          inspection.strip << ">"
        end
      end
    end

    private def each_inspectable
      inspectables = if __inspectables.any?
        __inspectables
      else
        instance_variables.map(&:to_s).reject { |instance_variable|
          instance_variable.start_with?("@_")
        }
      end

      inspectables.each do |inspectable|
        unless __uninspectables.include?(inspectable)
          yield inspectable
        end
      end
    end

    class << self
      include Is::Localized

      def prevent_recursion(object)
        object_id = object.object_id
        inspected_objects[object_id] = true
        yield
      ensure
        inspected_objects.delete(object_id)
      end

      def inspecting?(object)
        inspected_objects[object.object_id]
      end

      def inspected_objects
        localized(:__corerb_inspected_objects) || localize(:__corerb_inspected_objects, {})
      end
    end
  end
end
